# NOTE: this tests expect the Open LISA Server being run in TCP mode
# at port 8080 and with ENV=test
# Example: python main.py  --env test --mode TCP --tcp_port 8080 --log-level INFO
import json

import pytest

from ..domain.exceptions.could_not_connect_to_server import CouldNotConnectToServerException

from ..domain.exceptions.sdk_exception import OpenLISAException
from .. import SDK

LOCALHOST = "127.0.0.1"
SERVER_PORT = 8080
LOG_LEVEL = "ERROR"
TEST_TEKTRONIX_OSC_PHYSICAL_ADDRESS = "USB0::0x0699::0x0363::C107676::INSTR"
TEST_TEKTRONIX_OSC_INSTRUMENT_ID = 1
MOCK_CAMERA_ID = 2
UNEXISTING_INSTRUMENT_ID = 999
SOME_VALID_TEKTRONIX_OSC_COMMAND = "clear_status"
SOME_VALID_TEKTRONIX_OSC_COMMAND_INVOCATION = "clear_status"
MOCK_IMAGE_PATH = "Open_LISA_SDK/tests/mock_img.jpg"

VALID_INSTRUMENT_CREATION_DICT = {
    "brand": "some brand",
    "model": "some model",
    "physical_address": None,
    "type": "CLIB",
    "description": "some description"
}
VALID_UPDATED_BRAND = "some new brand"
VALID_INSTRUMENT_UPDATE_DICT = {
    "brand": VALID_UPDATED_BRAND
}


def test_get_instruments_as_python_list_of_dicts():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    instruments = sdk.get_instruments(response_format="PYTHON")
    assert isinstance(instruments, list)
    for i in instruments:
        assert isinstance(i, dict)
    sdk.disconnect()


def test_get_instruments_as_json_string():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    instruments_json_string = sdk.get_instruments(response_format="JSON")
    assert isinstance(instruments_json_string, str)

    try:
        json.loads(instruments_json_string)
    except:
        pytest.fail("invalid json string received: {}".format(
            instruments_json_string))

    sdk.disconnect()


def test_lock_is_released_after_exceptions():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    with pytest.raises(OpenLISAException):
        sdk.get_instrument(instrument_id=UNEXISTING_INSTRUMENT_ID,
                           response_format="PYTHON")
    with pytest.raises(OpenLISAException):
        sdk.get_instrument(instrument_id=UNEXISTING_INSTRUMENT_ID,
                           response_format="PYTHON")


def test_get_specific_instrument_as_python_dict():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    instrument = sdk.get_instrument(
        instrument_id=TEST_TEKTRONIX_OSC_INSTRUMENT_ID, response_format="PYTHON")
    assert isinstance(instrument, dict)
    assert instrument["physical_address"] == TEST_TEKTRONIX_OSC_PHYSICAL_ADDRESS

    sdk.disconnect()


def test_get_specific_instrument_as_json_str():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    instrument_json = sdk.get_instrument(
        instrument_id=TEST_TEKTRONIX_OSC_INSTRUMENT_ID, response_format="JSON")
    assert isinstance(instrument_json, str)
    assert TEST_TEKTRONIX_OSC_PHYSICAL_ADDRESS in instrument_json

    sdk.disconnect()


def test_get_specific_instrument_that_does_not_exist_raises_exception():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)

    with pytest.raises(OpenLISAException):
        sdk.get_instrument(instrument_id=UNEXISTING_INSTRUMENT_ID,
                           response_format="PYTHON")
    sdk.disconnect()


def test_get_instrument_commands_as_python_list_of_dicts():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    commands = sdk.get_instrument_commands(
        instrument_id=TEST_TEKTRONIX_OSC_INSTRUMENT_ID, response_format="PYTHON")
    assert isinstance(commands, dict)
    assert SOME_VALID_TEKTRONIX_OSC_COMMAND in commands.keys()
    sdk.disconnect()


def test_get_instrument_commands_as_json_string():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    commands_json_string = sdk.get_instrument_commands(
        instrument_id=TEST_TEKTRONIX_OSC_INSTRUMENT_ID, response_format="JSON")
    assert isinstance(commands_json_string, str)
    assert SOME_VALID_TEKTRONIX_OSC_COMMAND in commands_json_string
    sdk.disconnect()


def test_is_valid_command_invocation_with_a_valid_invocation_returns_true():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    is_valid = sdk.is_valid_command_invocation(
        instrument_id=TEST_TEKTRONIX_OSC_INSTRUMENT_ID,
        command_invocation=SOME_VALID_TEKTRONIX_OSC_COMMAND_INVOCATION)
    assert is_valid == True
    sdk.disconnect()


def test_is_valid_command_invocation_with_invalid_invocations_returns_false():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    invalid_invocations = [
        "unexisting command",
        "set_trigger_level 10 20",
        "set_trigger_level ASCII",
    ]
    for ii in invalid_invocations:
        is_valid = sdk.is_valid_command_invocation(
            instrument_id=TEST_TEKTRONIX_OSC_INSTRUMENT_ID,
            command_invocation=ii)
        print(ii)
        assert is_valid == False
    sdk.disconnect()


def test_send_command_to_get_image_from_mock_camera():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    result = sdk.send_command(instrument_id=MOCK_CAMERA_ID,
                              command_invocation="get_image")
    with open(MOCK_IMAGE_PATH, "rb") as f:
        image_bytes = f.read()
        assert image_bytes == result["value"]


def test_send_command_and_save_result_in_server():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    result = sdk.send_command(instrument_id=MOCK_CAMERA_ID,
                              command_invocation="get_image", command_result_output_file="sandbox/output.jpg")
    assert result == None
    sandbox_dir = sdk.get_directory_structure(
        "sandbox", response_format="PYTHON")
    # TODO: assert here that output.jpg exists in server
    sdk.delete_file(file_path="sandbox/output.jpg")
    sandbox_dir = sdk.get_directory_structure(
        "sandbox", response_format="PYTHON")
    # TODO: assert here that output.jpg does not exist in server


def test_instrument_CRUDs():
    sdk = SDK(log_level=LOG_LEVEL)
    sdk.connect_through_TCP(host=LOCALHOST, port=SERVER_PORT)
    new_instrument = sdk.create_instrument(
        new_instrument=VALID_INSTRUMENT_CREATION_DICT, response_format="PYTHON")
    assert new_instrument["brand"] == VALID_INSTRUMENT_CREATION_DICT["brand"]

    updated_instrument = sdk.update_instrument(
        instrument_id=new_instrument["id"], updated_instrument=VALID_INSTRUMENT_UPDATE_DICT, response_format="PYTHON")
    assert new_instrument["id"] == updated_instrument["id"]
    assert updated_instrument["brand"] == VALID_INSTRUMENT_UPDATE_DICT["brand"]

    deleted_instrument = sdk.delete_instrument(
        instrument_id=new_instrument["id"], response_format="PYTHON")
    assert new_instrument["id"] == deleted_instrument["id"]

    with pytest.raises(OpenLISAException):
        sdk.get_instrument(instrument_id=deleted_instrument["id"])

    sdk.disconnect()



import struct
import time
from .message_protocol import MessageProtocol
from ...logging import log


class MessageProtocolRS232(MessageProtocol):
    def __init__(self, rs232_connection):
        super().__init__(protocol_name="SERIAL")
        self._connection = rs232_connection
        if not self._connection.isOpen():
            self._connection.open()

    def disconnect(self):
        while self._connection.in_waiting:
            log.debug(
                '[MessageProtocolRS232][disconnect] waiting input buffer to be empty')
            time.sleep(0.5)
        self._connection.reset_input_buffer()

        while self._connection.out_waiting:
            log.debug(
                '[MessageProtocolRS232][disconnect] waiting output buffer to be empty')
            time.sleep(0.5)
        self._connection.reset_output_buffer()

        self._connection.close()

    def send_msg(self, msg, encode=True):
        if encode:
            msg = msg.encode()
        # Prefix each message with a 4-byte length
        msg = struct.pack('>I', len(msg)) + msg
        self._connection.write(msg)
        self._sent_bytes_amount_track += self.MSG_LEN_BYTES_AMOUNT + len(msg)

    def receive_msg(self, decode=True):
        # Read message length and unpack it into an integer
        raw_msglen = self.__recvall(self.MSG_LEN_BYTES_AMOUNT)
        self._received_bytes_amount_track += self.MSG_LEN_BYTES_AMOUNT
        if not raw_msglen:
            raise ConnectionResetError
        msglen = struct.unpack('>I', raw_msglen)[0]
        # Read the message data
        data = self.__recvall(msglen)
        self._received_bytes_amount_track += msglen

        if decode:
            data = data.decode()
        return data

    def __recvall(self, n):
        # Helper function to recv n bytes or raise ConnectionResetError if EOF is hit
        data = bytearray()
        while len(data) < n:
            bytes_to_read = max(
                1, min(2048, self._connection.in_waiting, n - len(data)))
            packet = self._connection.read(bytes_to_read)
            if not packet:
                raise ConnectionResetError
            data.extend(packet)
        return data



import socket
import struct
from .message_protocol import MessageProtocol


# TODO: Duplicado en server y SDK, ver de usar uno en comun
# Source: https://stackoverflow.com/questions/17667903/python-socket-receive-large-amount-of-data
class MessageProtocolTCP(MessageProtocol):
    def __init__(self, tcp_socket):
        super().__init__(protocol_name="TCP")
        self._socket = tcp_socket

    def disconnect(self):
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
        except (socket.error, OSError, ValueError):
            pass
        self._socket.close()

    def send_msg(self, msg, encode=True):
        if encode:
            msg = msg.encode()
        # Prefix each message with a 4-byte length (network byte order)
        msg = struct.pack('>I', len(msg)) + msg
        self._socket.sendall(msg)
        self._sent_bytes_amount_track += self.MSG_LEN_BYTES_AMOUNT + len(msg)

    def receive_msg(self, decode=True):
        # Read message length and unpack it into an integer
        raw_msglen = self.__recvall(self.MSG_LEN_BYTES_AMOUNT)
        self._received_bytes_amount_track += self.MSG_LEN_BYTES_AMOUNT
        if not raw_msglen:
            raise ConnectionResetError
        msglen = struct.unpack('>I', raw_msglen)[0]
        # Read the message data
        data = self.__recvall(msglen)
        self._received_bytes_amount_track += msglen
        if decode:
            data = data.decode()
        return data

    def __recvall(self, n):
        # Helper function to recv n bytes or raise ConnectionResetError if EOF is hit
        data = bytearray()
        while len(data) < n:
            packet = self._socket.recv(n - len(data))
            if not packet:
                raise ConnectionResetError
            data.extend(packet)
        return data




from pathlib import Path
from setuptools import setup, find_packages

VERSION_PLACEHOLDER = '0.9.9'

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='Open_LISA_SDK',         # How you named your package folder (MyLib)
    packages=find_packages(),
    # Start with a small number and increase it with every change you make
    version=VERSION_PLACEHOLDER,
    # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    license='MIT',
    # Give a short description about your library
    description='SDK for Laboratory Instrument Station Adapter',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Ariel Alvarez Windey & Gabriel Robles',                   # Type in your name
    author_email='ajalvarez@fi.uba.ar',      # Type in your E-Mail
    # Provide either the link to your github or to your website
    url='https://github.com/aalvarezwindey/Open-LISA-SDK',
    readme='https://github.com/aalvarezwindey/Open-LISA-SDK/README.md',
    download_url='https://github.com/aalvarezwindey/Open-LISA-SDK/archive/refs/tags/{}.tar.gz'.format(
        VERSION_PLACEHOLDER),
    keywords=['SDK', 'ELECTRONIC', 'INSTRUMENT', 'ADAPTER', 'FIUBA', 'OPEN',
              'LISA', 'LABORATORY'],   # Keywords that define your package best
    install_requires=["pyserial"],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 4 - Beta',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.6',
    ],
)
